/*
 * A simple Pong game for GP2X
 * (C) 2006 Vin Helminen
 * $Id: pong.c,v 1.5 2006-12-09 18:35:54 vah Exp $
 */

#include <stdio.h>
#include <stdlib.h>

#include <sys/time.h>
#include <unistd.h>

#include <SDL/SDL.h>

#include "gp2x.h"
#include "text.h"
#include "font.h"
#include "graphics.h"

#define BLOCK_SIZE 8  /* the diameter of the "ball" */
#define BAT_HEIGHT 48
#define BAT_WIDTH  6
#define BLOCK_V_SPEED 4
#define BLOCK_H_SPEED 5
#define PLAYER_SPEED 5
#define CPU_SPEED 4


#define PRINT_FPS 0

static inline void randomize(void)
{
  struct timeval tv;
  gettimeofday(&tv, NULL);
  srand((tv.tv_usec) | (tv.tv_sec));
}

static Uint32 randomColor(SDL_PixelFormat *format)
{
  return SDL_MapRGB(format, (0x70|rand())&0xff, (0x70|rand())&0xff, (0x70|rand())&0xff);
}

typedef struct { 
  int x, y, sx, sy;
} Object;

#define clearScreen(screen, color) fillRect(screen, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, color)

static inline void drawIntro(SDL_Surface *screen)
{
  Uint32 color;

  color = SDL_MapRGB(screen->format, 160, 160, 200);
  alignString("P O N G", CENTER, 1, 2*FONT_HEIGHT+1, screen, color);
  color = SDL_MapRGB(screen->format, 200, 200, 255);
  alignString("P O N G", CENTER, -1, 2*FONT_HEIGHT-1, screen, color);
  color = SDL_MapRGB(screen->format, 60, 60, 75);
  alignString("P O N G", CENTER, 0, 2*FONT_HEIGHT, screen, color);

  alignString("v0.0", RIGHT, -BAT_WIDTH, SCREEN_HEIGHT - FONT_HEIGHT, screen, color);

  color = SDL_MapRGB(screen->format, 100, 100, 160);
  alignString("for GP2X", CENTER, 0, 4*FONT_HEIGHT, screen, color);

  color = SDL_MapRGB(screen->format, 127, 250, 200);
  alignString(" 2006 Vin Helminen", CENTER, 0, 7*FONT_HEIGHT, screen, color);

  color = SDL_MapRGB(screen->format, 200, 155, 155);
  drawText("Buttons:\n\n \x95 (A)/(B)\tLaunch ball\n\t\tPower shot\n \x95 Up/Down\tMove bat\n\n \x95 Select\tReset game\n\n \x95 Start\tExit",
           SCREEN_WIDTH/6, SCREEN_HEIGHT/2+4*FONT_HEIGHT, screen, color);
}

int main(int argc, char **argv)
{
  SDL_Surface *screen;
  SDL_Joystick *joystick;
  SDL_Event event;
  Uint32 background, blockcolor, batcolor, textcolor;
  Object block, batl, batr; /* "ball" and left and right bats */
  int scorel = 0, scorer = 0;
  char buffer[256];
  int intro;
  enum { LEFT = 0, RIGHT } lastscorer;
  int DEBUG = argc > 1;
#if PRINT_FPS
  Uint32 ticks, oldticks = 0, dticks, frames = 0;
#endif
  if(DEBUG) puts("DEBUG");

  GP2XInit(!DEBUG);

  if(SDL_Init(SDL_INIT_VIDEO | SDL_INIT_JOYSTICK) < 0)
    return EXIT_FAILURE;

  screen = initGraphics();
  if(!screen) {
    SDL_Quit();
    return EXIT_FAILURE;
  }

  joystick = SDL_JoystickOpen(0);
  if(!joystick) {
    SDL_FreeSurface(screen);
    SDL_Quit();
    return EXIT_FAILURE;
  }

  randomize();

  background = SDL_MapRGB(screen->format, 0, 0, 0);
  batcolor = SDL_MapRGB(screen->format, 200, 200, 200);
  blockcolor = randomColor(screen->format);
  textcolor = SDL_MapRGB(screen->format, 255, 255, 155);

  GP2XCpuFrequency(50);
  
 reset:
  intro = 1;
  lastscorer = LEFT;
  clearScreen(screen, background);
  drawIntro(screen);
  
  /* init block */
  block.x = (SCREEN_WIDTH-BLOCK_SIZE)/2;
  block.y = (SCREEN_HEIGHT-BLOCK_SIZE)/2;
  block.sx = block.sy = 0;

  /* init left bat (=player) */
  batl.x = 0;
  batl.y = (SCREEN_HEIGHT-BAT_HEIGHT)/2;
  batl.sx = batl.sy = 0;

  /* init right bat (=computer) */
  batr.x = SCREEN_WIDTH-1-BAT_WIDTH;
  batr.y = (SCREEN_HEIGHT-BAT_HEIGHT)/2;
  batr.sx = batr.sy = 0;

  scorel = scorer = 0;

  /* game loop */
  while(1) {
    while(SDL_PollEvent(&event)) {
      switch(event.type) {
      case SDL_QUIT:
	goto exit;
	
      case SDL_JOYBUTTONDOWN:
	switch(event.jbutton.button) {
	  
	case GP2X_BUTTON_UP:
	  batl.sy = -PLAYER_SPEED;
	  break;
	case GP2X_BUTTON_DOWN:
	  batl.sy = PLAYER_SPEED;
	  break;
	  
	case GP2X_BUTTON_A:
	case GP2X_BUTTON_B:
	  if(block.sx == 0 && block.sy == 0) {
	    block.sy = rand()&(1<<9)?BLOCK_V_SPEED:-BLOCK_V_SPEED;
	    if(lastscorer == LEFT) {
	      block.sx = BLOCK_H_SPEED;
	      block.x = 0;
	      block.y = batl.y + (BAT_HEIGHT-BLOCK_SIZE)/2;
	    } else {
	      block.sx = -BLOCK_H_SPEED;
	      block.x = SCREEN_WIDTH-1 - BLOCK_SIZE;
	      block.y = batr.y + (BAT_HEIGHT-BLOCK_SIZE)/2;
	    }
	    
	    blockcolor = randomColor(screen->format); /* change block color for a visual effect */
	    
	    if(intro) {
	      scrollOut(screen, BAT_WIDTH, FONT_HEIGHT, SCREEN_WIDTH-1-2*BAT_WIDTH, SCREEN_HEIGHT-FONT_HEIGHT, background);
	      intro = 0;
	    }
	  } else if(block.x + block.sx < BAT_WIDTH) {
	    block.sx *= 2;
	    batl.sx = block.sx;
	  } else {
	    batl.sx = BLOCK_H_SPEED;
	  }
	  break;
	  
	case GP2X_BUTTON_START:
	  scrollOut(screen, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, background);
	  goto exit;
	  
	case GP2X_BUTTON_SELECT:
	  goto reset;
	  
	default:
	  break;
	} /* switch button */
	break;
	
      case SDL_JOYBUTTONUP:
	switch(event.jbutton.button) {
	case GP2X_BUTTON_UP:
	case GP2X_BUTTON_DOWN:
	  batl.sy = 0;
	  break;
	default:
	  break;
	} /* switch button */
	break;
	
      default:
	break;
      } /* switch event */
    } /* while pollevent */
    
    /* a very simple AI */
    batr.sy = 0;
    if(block.sx > 0) {
      int blocky = block.y + BLOCK_SIZE/2,
	  baty = batr.y + BAT_HEIGHT/2;
      if(blocky < baty)
        batr.sy = -CPU_SPEED;
      else if(blocky > baty)
        batr.sy = CPU_SPEED; 
    }
    
    if(block.sx || block.sy) {
      /* update screen */
      clearScreen(screen, background);
      
      /* right bat, cpu */
      batr.y += batr.sy;
      if(batr.y < 0) batr.y = 0;
      else if(batr.y >= SCREEN_HEIGHT-BAT_HEIGHT) batr.y = SCREEN_HEIGHT-1-BAT_HEIGHT;
      if(batr.sx) {
        batr.x += batr.sx;
        batr.sx = -batr.sx / 2;
      } else if(batr.x != SCREEN_WIDTH-1-BAT_WIDTH) {
        batr.sx = -batr.x+(SCREEN_WIDTH-1-BAT_WIDTH);
      }
      
      /* block/ball */
      block.y += block.sy;
      if(block.y < BLOCK_V_SPEED || block.y > SCREEN_HEIGHT-1-BLOCK_SIZE-BLOCK_V_SPEED) block.sy = -block.sy;   
      
      block.x += block.sx;
      if(block.x < 0) {
        block.x = 0;
        if(block.y+BLOCK_SIZE < batl.y || block.y > batl.y + BAT_HEIGHT) {
          block.sx = block.sy = 0;
          scorer++; /* cpu scores a point */
          lastscorer = RIGHT;
        } else {
          batl.sx = block.sx/2;
          block.sx = -block.sx;
        }
      } else if(block.x > SCREEN_WIDTH-1-BLOCK_SIZE) {
        block.x = SCREEN_WIDTH-1-BLOCK_SIZE;
        if(block.y+BLOCK_SIZE < batr.y || block.y > batr.y + BAT_HEIGHT) {
          block.sx = block.sy = 0; 
          scorel++; /* player scores a point */
          lastscorer = LEFT;
        } else {
          batr.sx = block.sx/2;
          block.sx = -block.sx;
        }
      } 
    } else {
      fillRect(screen, batl.x, batl.y, BAT_WIDTH, BAT_HEIGHT, background);
    }
    
    /* left bat, player */
    batl.y += batl.sy;
    if(batl.y < 0) batl.y = 0;
    else if(batl.y >= SCREEN_HEIGHT-BAT_HEIGHT) batl.y = SCREEN_HEIGHT-1-BAT_HEIGHT;
    if(batl.sx) {
      batl.x += batl.sx;
      batl.sx = -batl.sx / 2;
    } else if(batl.x != 0) {
      batl.sx = -batl.x;
    }
    fillRect(screen, batl.x, batl.y, BAT_WIDTH, BAT_HEIGHT, batcolor);
    
    /* paint right bat */
    fillRect(screen, batr.x, batr.y, BAT_WIDTH, BAT_HEIGHT, batcolor);
    
    /* paint block */
    fillRect(screen, block.x, block.y, BLOCK_SIZE, BLOCK_SIZE, blockcolor);
    
    /* paint scores */
    sprintf(buffer, "Player: %3d", scorel);
    alignString(buffer, LEFT, 1+BAT_WIDTH, 0, screen, textcolor);
    sprintf(buffer, "CPU: %3d", scorer);
    alignString(buffer, RIGHT, -1-BAT_WIDTH, 0, screen, textcolor);
    
    /* show changes */
    GP2XWaitVBlank();
    SDL_Flip(screen);
    
    
#if PRINT_FPS
    /* FPS counter */
    if(DEBUG) {
      ++frames;
      ticks = SDL_GetTicks();
      dticks = ticks - oldticks;
      if(dticks > 1000) {
	Uint32 i = (10000*frames)/dticks;
	printf("%d.%d FPS\r", i/10, i%10);
	fflush(stdout);
      }
      frames = 0; oldticks = ticks;
    }
#endif
  } /* while 1 */
  
 exit:
  SDL_FreeSurface(screen);
  SDL_Quit();
  
  return EXIT_SUCCESS;;
}
